#!/usr/bin/env python

#
# CodeGen.py
#
# Converts a parsed forest of 'ParseElement' into files (data & text)
#
#
# Copyright (C) 2007  Sylvain Munaut <tnt@246tNt.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

import Parser
import ParseElement as pe


class MemorySpace:

	def __init__(self):
		self.sections = []
		self.len = -1

	def appendSection(self, section):
		self.sections.append(section)

	def resolveSections(self, symtab = None):
		p = 0;
		for s in self.sections:
			p += s.resolve(p, symtab)
		self.len = p
		return p

	def generateImage(self):
		if self.len == -1:
			print "Trying to generate image without having resolvedSections"
			return None

		img = [0 for i in range(0, self.len)]

		for s in self.sections:
			s.fillImage(img)

		return img


class Section:

	CODE			= 1
	DATA			= 2
	INIT			= 4

	def __init__(self, name, flags):
		self.name = name
		self.flags = flags
		self.position = 0
		self.zones = []
		self.cur_zone = SectionZone(0)

	def getPosition(self):
		return pe.BinaryOp('+', pe.Symbol('__%s_zone%d' % (self.name, len(self.zones))), pe.Literal(self.cur_zone.len))

	def addEntry(self, entry, reloc = None):
		self.cur_zone.addEntry(entry, reloc)

	def directiveOrg(self, org):
		self.zones.append(self.cur_zone)
		self.cur_zone = SectionZone(origin=org)

	def directiveAlign(self, align):
		self.zones.append(self.cur_zone)
		self.cur_zone = SectionZone(align=align)

	def directiveSpace(self, space):
		base = pe.BinaryOp('+', pe.Symbol('__%s_zone%d_rel' % (self.name, len(self.zones))), pe.Literal(self.cur_zone.len))
		if type(space) == int:
			space = pe.Literal(space)
		org = pe.BinaryOp('+', base, space)

		self.zones.append(self.cur_zone)
		self.cur_zone = SectionZone(origin=org)

	def finish(self):
		self.zones.append(self.cur_zone)
		self.cur_zone = None

	def resolve(self, abs_origin, symtab):

		# Resolve loop init
		pos = 0
		idx = 0

		# Add begin symbol
		symtab['__%s_begin' % self.name] = pe.Literal(abs_origin)

		# Process zones
		for sz in self.zones:
			# Where to start
			org = sz.origin
			if not org is None:
				if type(org) == int:
					if org < pos:
						print "Attempt to go back in section %s. Ignored" % self.name
					else:
						pos = org
				else:
					if not org.evaluable(symtab):
						# Try harder
						org = org.resolve_syms(symtab).expand().collect()
					if not org.evaluable(symtab):
						print "Unable to resolve origin. That's bad ... Ignoring."
						print sz.origin
					else:
						pos = org.eval(symtab)

			# Alignement
			# FIXME not supported ...

			# Add zone begin symbol
			symtab['__%s_zone%d' % (self.name, idx)] = pe.Literal(abs_origin + pos)
			symtab['__%s_zone%d_rel' % (self.name, idx)] = pe.Literal(pos)

			# Save the final pos in the SectionZone object
			sz.abs_pos = abs_origin + pos

			# Increment of the size of this zone
			pos += sz.len

			# Next zone
			idx += 1

		# Add end symbol
		symtab['__%s_end' % self.name] = pe.Literal(abs_origin + pos)

		# Return length
		return pos

	def processRelocations(self, symtab = None):
		for sz in self.zones:
			sz.processRelocations(symtab)

	def fillImage(self, img):
		for sz in self.zones:
			img[sz.abs_pos:sz.abs_pos+sz.len] = sz.content


class SectionZone:

	def __init__(self, origin = None, align = 1):
			# The origin is relative to the section start
			# and can be : - None       -> no constraint
			#              - integer    -> absolute position in the section
			#              - Value obj. -> Value to be resolved later
		self.origin = origin
		self.content = []
		self.reloc = []
		self.len = 0

	def addEntry(self, entry, reloc = None):
		self.content.append(entry)
		if reloc:
			self.reloc.append( (self.len, reloc) )
		self.len += 1

	def processRelocations(self, symtab = None):
		for (p, r) in self.reloc:
			self.content[p] = r.execute(self.content[p], symtab)


class Relocation:

	def __init__(self, value, mask, shift):
		self.value = value
		self.mask  = mask
		self.shift = shift

	def execute(self, val, symtab = None):

		rv = self.value
		if not rv.evaluable(symtab):
			rv = rv.resolve_syms(symtab).expand().collect()
			if not rv.evaluable(symtab):
				print "Error during relocation. This is very bad. '%s'" % self.value
				return val
		rv = rv.eval(symtab)

		if self.shift < 0:
			rv >>= -self.shift
		else:
			rv << self.shift

		val |= rv & self.mask

		return val


# ----------------------------------------------------------------------------
# Main CodeGenerator class
# ----------------------------------------------------------------------------

	# Decorators
	# ----------
	# for the dir_ and ins_ handlers in CodeGen

def check_args_count(min, max):
	def fn_check_args_count(fn):
		def anon(*args, **kargs):
			l = args[1].args and len(args[1].args) or 0
			if (l < min):
				args[0].error("Missing arguments for '%s'" % args[1].name);
				return False
			if ((max >= 0) and (l > max)):
				args[0].error("Too much arguments for '%s'" % args[1].name);
				return False
			return fn(*args, **kargs)
		return anon
	return fn_check_args_count


def check_section_flags(flags):
	def fn_check_section_flags(fn):
		def anon(*args, **kargs):
			if ((args[0].cur_section.flags & flags) != flags):
				args[0].error("Invalid section for %s" % args[1]);
				return False
			return fn(*args, **kargs)
		return anon
	return fn_check_section_flags


	# Class code
	# ----------

class CodeGen:

	def __init__(self, sections):

		# Build section dictionnary
		self.sections = { }
		for s in sections:
			self.sections[s.name] = s

		self.cur_section = None

		# Symbol table
		self.symtab = { }


	def codegen(self, slist):

		for s in slist:
			if isinstance(s, pe.Directive):
				dfn = self.__class__.__dict__.get("dir_" + s.name)
				if not dfn and self.sections.has_key(s.name):
					dfn = self.__class__.__dict__.get("dir_section_direct")
				if not dfn:
					self.error("Unknown directive '%s'" % s.name)
					continue
				dfn(self, s)
			elif isinstance(s, pe.Instruction):
				ifn = self.__class__.__dict__.get("ins_" + s.name)
				if not ifn:
					self.error("Unknown instruction '%s'" % s.name)
					continue
				ifn(self, s)
			elif isinstance(s, pe.Label):
				self.setSymbol(s.name, self.cur_section.getPosition())


	def switch_section(self, section_name):
		ns = self.sections.get(section_name)
		if not ns:
			self.error("Unknown section '%s'" % section_name)
			return
		self.cur_section = ns

	def setSymbol(self, name, value):
		if self.symtab.has_key(name):
			print "Duplicate symbol definition '%s'" % name
			return
		self.symtab[name] = value


	def error(self, err_txt):
		print err_txt


	#
	# Directives
	#

	@check_args_count(1,1)
	def dir_section(self, directive):
		self.switch_section(directive.args[0])

	@check_args_count(0,0)
	def dir_section_direct(self, directive):
		self.switch_section(directive.name)

	@check_args_count(1,1)
	def dir_org(self, directive):
		org = directive.args[0]
		if org.evaluable(self.symtab):
			org = org.eval(self.symtab)
		self.cur_section.directiveOrg(org)

	@check_args_count(1,1)
	def dir_align(self, directive):
		if not directive.args[0].evaluable(self.symtab):
			print "ALIGN directive with non absolute argument. Ignored."
			return
		self.cur_section.directiveAlign(directive.args[0].eval(self.symtab))

	@check_args_count(1,-1)
	def dir_space(self, directive):
		for s in directive.args:
			if s.evaluable(self.symtab):
				s = s.eval(self.symtab)
			self.cur_section.directiveSpace(s)

	@check_section_flags(Section.DATA | Section.INIT)
	@check_args_count(1,-1)
	def dir_ascii(self, directive):
		for s in directive.args:
			if not isinstance(s, pe.QuotedString):
				print "Invalid argument"
				continue
			for c in s.value:
				self.cur_section.addEntry(ord(c) & 0xff)

	@check_section_flags(Section.DATA | Section.INIT)
	@check_args_count(1,-1)
	def dir_asciz(self, directive):
		for s in directive.args:
			if not isinstance(s, pe.QuotedString):
				print "Invalid argument"
				continue
			for c in s.value:
				self.cur_section.addEntry(ord(c) & 0xff)
			self.cur_section.addEntry(0)

	@check_section_flags(Section.DATA | Section.INIT)
	@check_args_count(1,-1)
	def dir_byte(self, directive):
		for b in directive.args:
			if not isinstance(b, pe.Value):
				print "Invalid argument"
				continue
			if b.evaluable(self.symtab):
				b = b.eval(self.symtab)
				self.cur_section.addEntry(b & 0xff)
			else:
				self.cur_section.addEntry(0)
				print "TODO: Relocation for %s" % b

	@check_section_flags(Section.DATA | Section.INIT)
	@check_args_count(1,-1)
	def dir_hword(self, directive):
		for b in directive.args:
			if not isinstance(b, pe.Value):
				print "Invalid argument"
				continue
			if b.evaluable(self.symtab):
				b = b.eval(self.symtab)
				self.cur_section.addEntry((b >> 8) & 0xff)
				self.cur_section.addEntry( b       & 0xff)
			else:
				self.cur_section.addEntry(0)
				self.cur_section.addEntry(0)
				print "TODO: Relocation for %s" % b

	@check_section_flags(Section.DATA | Section.INIT)
	@check_args_count(1,-1)
	def dir_word(self, directive):
		for b in directive.args:
			if not isinstance(b, pe.Value):
				print "Invalid argument"
				continue
			if b.evaluable(self.symtab):
				b = b.eval(self.symtab)
				self.cur_section.addEntry((b >> 24) & 0xff)
				self.cur_section.addEntry((b >> 16) & 0xff)
				self.cur_section.addEntry((b >> 8)  & 0xff)
				self.cur_section.addEntry( b        & 0xff)
			else:
				self.cur_section.addEntry(0)
				self.cur_section.addEntry(0)
				self.cur_section.addEntry(0)
				self.cur_section.addEntry(0)
				print "TODO: Relocation for %s" % b

	@check_args_count(2,2)
	def dir_equ(self, directive):
		n = directive.args[0]
		v = directive.args[1]
		if not isinstance(n, pe.Identifier):
			self.error("Invalid first argument for EQU")
			return
		self.setSymbol(n.name, v)

	dir_set = dir_equ


	#
	# Instructions
	#

	@check_section_flags(Section.CODE)
	@check_args_count(2,2)
	def ins_fmt1(self, instruction, base):
		# add, addcy, and, compare, load, or, sub, subcy, test, xor
		#  First arg is a register
		#  Second arg is either a register or a 8 bit const
		if not isinstance(instruction.args[0], pe.Identifier):
			print "Invalid argument"
			return

		r = instruction.args[0].getRegister(self.symtab)
		if r < 0:
			print "Unable to resolve symbol '%s' to a register" % instruction.args[0].name
			return
		base |= (r << 8)

		reloc = None
		if isinstance(instruction.args[1], pe.Identifier) and \
		   (instruction.args[1].getRegister(self.symtab) >= 0):
			r = instruction.args[1].getRegister(self.symtab)
			base |= (r << 4) | 0x01000
		elif isinstance(instruction.args[1], pe.Value):
			if instruction.args[1].evaluable(self.symtab):
				base |= (instruction.args[1].eval(self.symtab) & 0xff)
			else:
				reloc = Relocation(instruction.args[1], 0xff, 0)
		else:
			print "Invalid argument"
			return

		self.cur_section.addEntry(base, reloc)

	@check_section_flags(Section.CODE)
	@check_args_count(1,1)
	def ins_fmt2(self, instruction, base):
		# rl, rr, sl0, sl1, sla, slx, sr0, sr1, sra, srx
		#  First arg is a register
		if not isinstance(instruction.args[0], pe.Identifier):
			print "Invalid argument"
			return

		r = instruction.args[0].getRegister(self.symtab)
		if r < 0:
			print "Unable to resolve symbol '%s' to a register" % instruction.args[0].name
			return

		self.cur_section.addEntry(base | (r<<8))

	@check_section_flags(Section.CODE)
	@check_args_count(1,2)
	def ins_fmt3(self, instruction, base):
		# call, jump
		#  First arg is either address or condition code
		#  Second arg is an address (only if first was cc)
		if len(instruction.args) == 2:
			cc = instruction.args[0].name.lower()
			dst = instruction.args[1]
			if not self.cc_mask.has_key(cc):
				print "Invalid condition code"
			else:
				base |= self.cc_mask[cc]
		else:
			dst = instruction.args[0]

		reloc = None
		if dst.evaluable(self.symtab):
			base |= dst.eval(symtab) & 0x3ff
		else:
			reloc = Relocation(dst, 0x3ff, 0)

		self.cur_section.addEntry(base, reloc)

	@check_section_flags(Section.CODE)
	@check_args_count(2,2)
	def ins_fmt4(self, instruction, base):
		# input, output, store, fetch
		#  First argument is a register
		#  Second arg is either a constand address or a
		#   register + offset
		if not isinstance(instruction.args[0], pe.Identifier):
			print "Invalid argument"
			return

		r = instruction.args[0].getRegister(self.symtab)
		if r < 0:
			print "Unable to resolve symbol '%s' to a register" % instruction.args[0].name
			return
		base |= (r << 8)

		reloc = None
		if isinstance(instruction.args[1], pe.IndirectAddress):
			r = instruction.args[1].reg.getRegister(self.symtab)
			if r < 0:
				print "Unable to resolve symbol '%s' to a register" % instruction.args[0].name
				return
			base |= (r << 4) | (instruction.args[1].ofs & 0xf) | 0x01000
		elif isinstance(instruction.args[1], pe.Value):
			if instruction.args[1].evaluable(self.symtab):
				base |= (instruction.args[1].eval(self.symtab) & 0xff)
			else:
				reloc = Relocation(instruction.args[1], 0xff, 0)
		else:
			print "Invalid argument"
			return

		self.cur_section.addEntry(base, reloc)

	@check_section_flags(Section.CODE)
	@check_args_count(0,0)
	def ins_fmt5(self, instruction, base):
		# cli, sti
		#  No argument. Those are fixed opcodes
		self.cur_section.addEntry(base)

	ins_add		= lambda i, s : i.ins_fmt1(s, 0x18000)
	ins_addcy	= lambda i, s : i.ins_fmt1(s, 0x1a000)
	ins_and		= lambda i, s : i.ins_fmt1(s, 0x0a000)
	ins_call	= lambda i, s : i.ins_fmt3(s, 0x30000)
	ins_compare	= lambda i, s : i.ins_fmt1(s, 0x14000)
	ins_cli		= lambda i, s : i.ins_fmt5(s, 0x3c000)
	ins_sti		= lambda i, s : i.ins_fmt5(s, 0x3c001)
	ins_fetch	= lambda i, s : i.ins_fmt4(s, 0x06000)
	ins_input	= lambda i, s : i.ins_fmt4(s, 0x04000)
	ins_jump	= lambda i, s : i.ins_fmt3(s, 0x34000)
	ins_load	= lambda i, s : i.ins_fmt1(s, 0x00000)
	ins_or		= lambda i, s : i.ins_fmt1(s, 0x0c000)
	ins_output	= lambda i, s : i.ins_fmt4(s, 0x2c000)
	ins_rl		= lambda i, s : i.ins_fmt2(s, 0x20002)
	ins_rr		= lambda i, s : i.ins_fmt2(s, 0x2000c)
	ins_sl0		= lambda i, s : i.ins_fmt2(s, 0x20006)
	ins_sl1		= lambda i, s : i.ins_fmt2(s, 0x20007)
	ins_sla		= lambda i, s : i.ins_fmt2(s, 0x20000)
	ins_slx		= lambda i, s : i.ins_fmt2(s, 0x20004)
	ins_sr0		= lambda i, s : i.ins_fmt2(s, 0x2000e)
	ins_sr1		= lambda i, s : i.ins_fmt2(s, 0x2000f)
	ins_sra		= lambda i, s : i.ins_fmt2(s, 0x20008)
	ins_srx		= lambda i, s : i.ins_fmt2(s, 0x2000a)
	ins_store	= lambda i, s : i.ins_fmt4(s, 0x2e000)
	ins_sub		= lambda i, s : i.ins_fmt1(s, 0x1c000)
	ins_subcy	= lambda i, s : i.ins_fmt1(s, 0x1e000)
	ins_test	= lambda i, s : i.ins_fmt1(s, 0x12000)
	ins_xor		= lambda i, s : i.ins_fmt1(s, 0x0e000)

	@check_section_flags(Section.CODE)
	@check_args_count(0,1)
	def ins_return(self, instruction):
		base = 0x2a000
		if instruction.args and len(instruction.args) == 1:
			cc = instruction.args[0].name.lower()
			if not self.cc_mask.has_key(cc):
				print "Invalid condition code for return"
			else:
				base |= self.cc_mask[cc]

		self.cur_section.addEntry(base)

	@check_section_flags(Section.CODE)
	@check_args_count(1,1)
	def ins_returni(self, instruction):
		mode = instruction.args[0].name.lower()
		if mode == "enable":
			self.cur_section.addEntry(0x38001)
		elif mode == "disable":
			self.cur_section.addEntry(0x38000)
		else:
			print "Invalid argument for returni"


	cc_mask = {
		'c'		:	0x01800,
		'nc'	:	0x01c00,
		'nz'	:	0x01400,
		'z'		:	0x01000,
	};


# ----------------------------------------------------------------------------
# Main function, for testing
# ----------------------------------------------------------------------------

# Main
def main(argv = None):

	# Arguments
	if argv is None:
		argv = sys.argv

	src_files = [ os.path.dirname(argv[0]) + '/boot.S' ]
	src_files.extend(argv[1:-1])

	# Create sections
	section_text	= Section('text',   Section.CODE | Section.INIT)
	section_rodata	= Section('rodata', Section.DATA | Section.INIT)
	section_data	= Section('data',   Section.DATA | Section.INIT)
	section_bss		= Section('bss',	Section.DATA)

	sections = [ section_text, section_rodata, section_data, section_bss ]

	# Construct parser and parse the files
	p = Parser.Parser()
	pf = [ ]
	for f in src_files:
		print "Parsing %s" % f
		pf.extend(p.parse_file(f))

	# Assign the sections to files
	codefile = MemorySpace()
	datafile = MemorySpace()

	codefile.appendSection(section_text)
	datafile.appendSection(section_data)
	datafile.appendSection(section_rodata)	# rodata must be after data, if not,
	datafile.appendSection(section_bss)		# sections won't resolve properly

	# Do the code generation step
	cg = CodeGen(sections)
	cg.codegen(pf)

	# Finish all section
	for s in sections:
		s.finish()

	# Resolve the sections (this makes all the symbols defined by assigning
	# 'real' address to each section)
	codefile.resolveSections(cg.symtab)
	datafile.resolveSections(cg.symtab)

	# Process the relocations in each sections
	for s in sections:
		s.processRelocations(cg.symtab)

	for n in sorted(cg.symtab.keys()):
		v = cg.symtab[n]
		t1 = v.evaluable(cg.symtab)
		t2 = v.resolve_syms(cg.symtab).expand().collect().evaluable(cg.symtab)
		rv = 0
		if t2:
			rv = v.resolve_syms(cg.symtab).expand().collect().eval(cg.symtab)
		print '%20s -> %5s %5s %5d %s' % (n, t1, t2, rv, v)

	# Get the images
	img_code = codefile.generateImage()
	img_data = datafile.generateImage()

	# Copy the .data section into the data_init zone in .rodata
	db = cg.symtab['__data_begin'].eval()
	de = cg.symtab['__data_end'].eval()
	dib = cg.symtab['__data_init_base'].eval(cg.symtab)
	img_data[dib:dib+(de-db)] = img_data[db:de]

	# Write file to disk
	of = open(argv[-1], "w")

		# Code
	of.write("@0\n")
	for i in img_code:
		of.write("%05x\n" % i)

		# Spacing
	if len(img_code) != 896:
		for i in range(0,(896 - len(img_code))):
			of.write("00000\n")

		# Data
	img_data.append(0)
	for i in range(0, len(img_data)-1, 2):
		lsb = img_data[i]
		msb = img_data[i+1]
		of.write("0%02x%02x\n" % (msb, lsb))

	of.close()

	print "--- Code (%d)" % len(img_code)
	print "--- Data (%d)" % len(img_data)

# Run it
if __name__ == '__main__':
	main()

