#!/usr/bin/env python

#
# ParseElement.py
#
# Collection of elements generated during the parsing
#
#
# Copyright (C) 2007  Sylvain Munaut <tnt@246tNt.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re


# ---------------------------------------------------------------------------
# First level elements
# ---------------------------------------------------------------------------

class Directive:

	def __init__(self, name, args):
		self.name = name
		self.args = args

	def __str__(self):
		return "Directive('%s', %s)" % (self.name, self.args)


class Instruction:

	def __init__(self, name, args):
		self.name = name
		self.args = args

	def __str__(self):
		return "Instruction('%s', %s)" % (self.name, self.args)


class Label:

	def __init__(self, name):
		self.name = name

	def __str__(self):
		return "Label('%s')" % self.name


class LocalLabel(Label):

	def __init__(self, id, ll_tbl):
		id = int(id)
		ln = 'L' + str(id) + '_POS' + str(ll_tbl[id])
		ll_tbl[id] += 1
		Label.__init__(self, ln)


# ---------------------------------------------------------------------------
# Argument elements
# ---------------------------------------------------------------------------

	# 'Value' ones
	# ------------
	# Those can potentially represent a value, so there is a lot
	# of support added to deal with that possibility and symbolic
	# expression handling

# Base class to handle default behavior
class Value:

	def eval(self, symtab = None):
		return 0

	def evaluable(self, symtab = None):
		return False

	def resolve_syms(self, symtab = None):
		return self

	def expand(self):
		return self

	def collect(self):
		return self

	def get_terms(self):
		if self.evaluable():
			return [ (self.eval(), None) ]
		else:
			return [ (1, self) ]

	def can_distribute(self):
		return False


class Literal(Value):

	def __init__(self, value):
		self.value = value

	def eval(self, symtab = None):
		return self.value

	def evaluable(self, symtab = None):
		return True

	def __str__(self):
		return "Literal(%d)" % self.value


class Symbol(Value):

	def __init__(self, name):
		self.name = name

	def eval(self, symtab = None):
		if not symtab or not symtab.has_key(self.name):
			print "Undefined symbol '%s', evaluated as zero" % self.name
			return 0
		return symtab[self.name].eval(symtab)

	def evaluable(self, symtab = None):
		return symtab and symtab.has_key(self.name) and symtab[self.name].evaluable(symtab) or False

	def resolve_syms(self, symtab = None):
		return symtab and symtab.has_key(self.name) and symtab[self.name] or self;

	def __str__(self):
		return "Symbol('%s')" % self.name




class BinaryOp(Value):

	def __init__(self, op, a, b):
		self.op = op
		self.a  = a
		self.b  = b

	def eval(self, symtab = None):
		if self.op == '+':
			return self.a.eval(symtab) + self.b.eval(symtab)
		elif self.op == '-':
			return self.a.eval(symtab) - self.b.eval(symtab)
		elif self.op == '*':
			return self.a.eval(symtab) * self.b.eval(symtab)
		elif self.op == '/':
			return self.a.eval(symtab) / self.b.eval(symtab)
		else:
			print "Invalid op in BinaryOp, evaluated as 0"
			return 0

	def evaluable(self, symtab = None):
		return self.a.evaluable(symtab) and self.b.evaluable(symtab) or False

	def resolve_syms(self, symtab = None):
		return BinaryOp(self.op, self.a.resolve_syms(symtab), self.b.resolve_syms(symtab))

	def expand(self):
		if self.op == '*':
			if self.a.can_distribute():
				return self.a.distribute(lambda x: BinaryOp(self.op,x,self.b)).expand()
			if self.b.can_distribute():
				return self.b.distribute(lambda x: BinaryOp(self.op,x,self.a)).expand()
		elif self.op == '/':
			if self.a.can_distribute():
				return self.a.distribute(lambda x: BinaryOp(self.op,x,self.b)).expand()

		ae = self.a.expand()
		if ae.evaluable():
			ae = Literal(ae.eval())

		be = self.b.expand()
		if be.evaluable():
			be = Literal(be.eval())

		return BinaryOp(self.op, ae, be)

	def collect(self):
		# Get the terms
		terms = self.get_terms()

		# Accumulate for each term
		number = 0
		symbols = dict()
		others = []

		for (c, v) in terms:
			if not v:
				number += c
			elif v.evaluable():
				number += c * v.eval()
			elif hasattr(v, 'name'):
				if symbols.has_key(v.name):
					symbols[v.name] += c
				else:
					symbols[v.name] = c
			else:
				others.append( (c,v) )

		# Create final expression
		rv = number and Literal(number)
		for (c, v) in others:
			if not c:
				continue
			v = (c == 1) and v or BinaryOp('*', v, Literal(c))
			rv = rv and BinaryOp('+', v, rv) or v
		for (n, c) in symbols.iteritems():
			if not c:
				continue
			v = (c == 1) and Symbol(n) or BinaryOp('*', Symbol(n), Literal(c))
			rv = rv and BinaryOp('+', v, rv) or v

		return rv or Literal(0)

	def can_distribute(self):
		return self.op == '+' or self.op == '-'

	def distribute(self, dist_func):
		return BinaryOp(self.op, dist_func(self.a), dist_func(self.b))

	def get_terms(self):
		if self.evaluable():
			return [ (self.eval(), None) ]
		elif self.op == '+':
			rv = self.a.get_terms()
			rv.extend(self.b.get_terms())
			return rv
		elif self.op == '-':
			rv = [ (-c,v) for (c,v) in self.b.get_terms() ]
			rv.extend(self.a.get_terms())
			return rv
		elif self.op == '*':
			if self.a.evaluable():
				return [ (self.a.eval(), self.b) ]
			elif self.b.evaluable():
				return [ (self.b.eval(), self.a) ]

		return [ (1, self) ]

	def __str__(self):
		return "BinaryOp('%s', %s, %s)" % (self.op, self.a, self.b)


	# Classic ones
	# ------------

class IndirectAddress:

	def __init__(self, reg, ofs):
		self.reg = reg
		self.ofs = ofs

	def __str__(self):
		return "IndirectAddress('%s', %d)" % (self.reg, self.ofs)


class QuotedString:

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "QuotedString('%s')" % self.value


class Identifier(Symbol):	# Derivate from symbol because it may identify a symbol value ...

	def __init__(self, name):
		Symbol.__init__(self, name)

	def getRegister(self, symtab):
		v = self.name
		while symtab.has_key(v):
			nv = symtab[v]
			if not isinstance(nv, Identifier):
				break
			v = nv.name
		m = re.match("^s([0-9]|1[0-5])$", v)
		if m:
			return int(m.group(1))
		else:
			return -1

	def __str__(self):
		return "Identifier('%s')" % self.name


class LocalIdentifier(Identifier):

	def __init__(self, id, ll_tbl):
		ofs = id[1] == 'b' and -1 or 0
		id = int(id[0])
		ln = 'L' + str(id) + '_POS' + str(ll_tbl[id] + ofs)
		Identifier.__init__(self, ln)

