#!/usr/bin/env python

#
# Parser.py
#
# Parser using ply.yacc
# Converts the token stream produced by the Lexer into a forest
# of 'ParseElement'. It defines the grammar of the assembler
#
#
# Copyright (C) 2007  Sylvain Munaut <tnt@246tNt.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ply.yacc as yacc
import Lexer

import ParseElement as pe


class Parser:

	def __init__(self):
		self.tokens	= Lexer.tokens
		self.local_label_counter = [0 for i in range(0,10)]


	def p_program_simple(self, p):
		'''program : statement'''
		p[0] = [ p[1] ]

	def p_program_recurse(self, p):
		'''program : program statement'''
		p[1].append(p[2])
		p[0] = p[1]


	def p_statement_label(self, p):
		'''statement	: LABEL
						| LABEL NEWLINE'''
		p[0] = pe.Label(p[1])

	def p_statement_label_local(self, p):
		'''statement	: LABEL_LOCAL
						| LABEL_LOCAL NEWLINE'''
		p[0] = pe.LocalLabel(p[1], self.local_label_counter)

	def p_statement_directive(self, p):
		'''statement	: DIRECTIVE arg_list NEWLINE'''
		p[0] = pe.Directive(p[1], p[2])

	def p_statement_instruction(self, p):
		'''statement	: INSTRUCTION arg_list NEWLINE'''
		p[0] = pe.Instruction(p[1], p[2])


	def p_arg_list_none(self, p):
		'arg_list : '
		p[0] = None

	def p_arg_list_simple(self, p):
		'arg_list : argument'
		p[0] = [ p[1] ]

	def p_arg_list_recurse(self, p):
		'arg_list : arg_list COMMA argument'
		p[1].append(p[3])
		p[0] = p[1]


	def p_argument_qstring(self, p):
		'argument : QSTRING'
		p[0] = pe.QuotedString(p[1])

	def p_argument_iaddr(self, p):
	 	'argument : LPAREN IDENTIFIER RPAREN'
		p[0] = pe.IndirectAddress(pe.Identifier(p[2]), 0)

	def p_argument_iaddroffs(self, p):
		'argument : LPAREN IDENTIFIER RPAREN NUMBER'
		p[0] = pe.IndirectAddress(pe.Identifier(p[2]), p[4])

	def p_argument_expr(self, p):
		'argument : expression'
		p[0] = p[1]


	def p_expression_binary(self, p):
		'''expression	: expression PLUS expression
						| expression MINUS expression
						| expression TIMES expression
						| expression DIVIDES expression'''
		p[0] = pe.BinaryOp(p[2], p[1], p[3])

	def p_expression_uminus(self, p):
		'expression : MINUS expression %prec UMINUS'
		p[0] = pe.BinaryOp('-', pe.Literal(0), p[2])

	def p_expression_paren(self, p):
		'expression : LPAREN expression RPAREN'
		p[0] = p[2]

	def p_expression_identifier(self, p):
		'expression : IDENTIFIER'
		p[0] = pe.Identifier(p[1])

	def p_expression_identifier_local(self, p):
		'expression : IDENTIFIER_LOCAL'
		p[0] = pe.LocalIdentifier(p[1], self.local_label_counter)

	def p_expression_number(self, p):
		'expression : NUMBER'
		p[0] = pe.Literal(p[1])


	precedence = (
			('left', 'PLUS', 'MINUS'),
			('left', 'TIMES', 'DIVIDES'),
			('right', 'UMINUS' ),
		)


	def p_error(self, t):
		print "Syntax error in input ! (at token %s, line %d))" % ( t.type, t.lineno )


	def parse_file(self, filename):
		asmlexer = Lexer.Lexer()
		asmlexer.build()
		parser = yacc.yacc(module=self)

		f = open(filename, 'r')
		text = f.read();
		f.close()

		return parser.parse(text, lexer=asmlexer.lexer)

